//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Target.cpp                                    =
//=  Version: 2.1                                           =
//=  Authors: Jeff Rosen and Andrew Carlson                 =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//===========================================================

#include "Target.h"
#include "GlobalParams.h"
#include "Winnow.h"
#include "Perceptron.h"
#include "NaiveBayes.h"
#include <math.h>
#include <iostream>
#include <iomanip>

extern int examples;

void Target::Discard()
{
    // NOTE: This routine is only for the DISCARD_ABS mode,
    //       do NOT call this routine unless
    //       discardMode == DISCARD_ABS. The DISCARD_REL mode
    //       should be handled by the Network.
    FeatureMap::iterator feat = features.begin();

    discardThreshold = pAlgorithm->defaultWeight * globalParams.discardThreshold;
    
    // Update the weight vector
    for ( ; feat != features.end(); feat++)
    {
        if (feat->second.weight < discardThreshold)
        {
            feat->second.discarded = discard;
            activeFeatureCount--;
        }
    }
}

void Target::Discard( FeatureID id, double delta )
{
    // NOTE: This routine is only for the DISCARD_REL mode,
    //       do NOT call this routine unless
    //       discardMode == DISCARD_REL. Unlike Discard above,
    //       this routine operates on the specified feature (id)
    //       only rather than processing entire weight vector.
    
    FeatureMap::iterator feat = features.find(id);

    // Update the feature weight
    if (feat != features.end())
    {
        if (feat->second.discarded == active)
        {
            feat->second.weight -= delta;
            if (feat->second.weight == 0.0)
            {
                feat->second.discarded = discard;
                activeFeatureCount--;
            }
        }    
    }
}

void Target::ShowStatistics( Counter total ) const
{
    cout << "Statistics for target (" << targetID
         << ")" << endl;
    
    cout << "   Active: " << activeCount << " / "
         << (activeCount + nonActiveCount) << endl;
}

void Target::ShowFeatures() const
{
    cout << endl << "Features for target (" << targetID
         << ")  log(prior) = " << priorProbability << endl;
    
    FeatureMap::const_iterator it = features.begin();
    FeatureMap::const_iterator end = features.end();

    for ( ; it != end; it++)
    {
        cout << "  " << it->first << " -> " << it->second.weight << endl;
    }
}

void Target::Read( ifstream& in , AlgorithmVector& algorithms)
{
    features.clear();

    // Read the target ID and prior Probability
    in >> targetID >> priorProbability >> confidence >> activeCount 
       >> activeFeatureCount;

    if (globalParams.targetIds.find(targetID) == globalParams.targetIds.end())
    {
        globalParams.targetIds.insert(targetID);
    } 

    // Create the algorithm object
    string networkType;
    char delim;
    int index;  

    operator>>(in, networkType);
    delim = in.get();
   
    in >> index; 

    // Check for an algorithm with the same index
    AlgorithmVector::iterator algIt = algorithms.begin();
    while ((algIt != algorithms.end()) && ((*algIt)->index != index)) 
    {
        algIt++;
    }
   
    // no algorithm with the current index exists, so create it
    if (algIt == algorithms.end()) 
    { 
        if (networkType == "winnow")
        {
            pAlgorithm = new Winnow;
        } else if (networkType == "perceptron")
        {
            pAlgorithm = new Perceptron;
        } else if (networkType == "naivebayes")
        {
            pAlgorithm = new NaiveBayes;
        }
        pAlgorithm->Read(in);
        pAlgorithm->index = index;
        algorithms.push_back(pAlgorithm);
    } else { 
        // an algorithm with the current index exists, so use that 
        // algorithm and skip the rest of the input line
        pAlgorithm = *algIt;
        char skipline[80];
        in.getline(skipline, 80);
    }

    if (in.fail())
    {
        in.clear();
        cerr << "Failed reading target!" << endl;
        return;
    }

    if (globalParams.verbosity >= VERBOSE_MED)
    {
        cout << endl << "Created target " << targetID << endl;
    }

    FeatureID id;
    double weight;
    unsigned long count;
    int skip;
    int updates;
    // use skip and delim to skip over unnecessary values and colons 
    in >> skip >> delim >> skip >> delim >> id >> delim >> count >> updates >> weight;
    
    while (!in.fail())
    {
        features.insert(make_pair(id, MinimalFeature(weight)));
        features.find(id)->second.updates = updates; 
        features.find(id)->second.activeCount = count; 
        // use skip and delim to skip over unnecessary values and colons 
        in >> skip >> delim >> skip >> delim 
           >> id >> delim >> count >> updates >> weight;
    }

    in.clear();
}

void Target::Write( ofstream& out )
{
   
   if (globalParams.verbosity >= VERBOSE_MED)
    {
        cout << endl << "Features for target (" << targetID
             << ")" << endl;
    }

    // Write the target label
#ifdef WIN32    
    out << endl;
    operator<<(out, "target ");
    out << targetID << ' '
        << setprecision(12) << priorProbability << ' '
        << setprecision(12) << confidence << ' ' 
        << activeCount << ' '
        << activeFeatureCount << ' ';
    pAlgorithm->Write(out);
#else
    out.clear();
    out << "target " << targetID << ' ';
    out.setf(ios::scientific);
    out << setprecision(12) << priorProbability << ' ';
    out.setf(ios::fixed);
    out << confidence << ' '
        << activeCount << ' ' 
        << activeFeatureCount << ' ';
    pAlgorithm->Write(out);
#endif
    // Write-out each feature and its parameters
    FeatureMap::const_iterator it = features.begin();
    FeatureMap::const_iterator end = features.end();
    for ( ; it != end; it++)
    {
        if (it->second.discarded == active)
        {   
            if (globalParams.verbosity >= VERBOSE_MED)
            {
                cout << " " << it->first << " -> "
                     << it->second.weight << endl;
            }   
        
            out << targetID << " : " 
                << pAlgorithm->index << " : " 
                << setw(10) << it->first << " : "
                << setw(10) << it->second.activeCount 
                << setw(10) << it->second.updates
                << setw(18) 
                << setprecision(10) << it->second.weight
                << endl;
        }
    }
}

